# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Some code snippets are taken from the X86InductorQuantizer tutorial.
# https://pytorch.org/tutorials/prototype/pt2e_quant_x86_inductor.html
"""The quantizer using PT2E path."""

from typing import Any

import torch
import torch.ao.quantization.quantizer.x86_inductor_quantizer as xiq
from torch.ao.quantization.quantize_pt2e import convert_pt2e, prepare_pt2e
from torch.ao.quantization.quantizer.x86_inductor_quantizer import X86InductorQuantizer
from torch.fx.graph_module import GraphModule

from neural_compressor.common.utils import logger
from neural_compressor.torch.algorithms.base_algorithm import Quantizer
from neural_compressor.torch.algorithms.pt2e_quant import half_precision_rewriter as hp_rewriter
from neural_compressor.torch.algorithms.pt2e_quant.utility import create_xiq_quantizer_from_pt2e_config


class W8A8PT2EQuantizer(Quantizer):
    """The W8A8 quantizer using PT2E."""

    is_dynamic = False

    def __init__(self, quant_config=None):
        """Initialize the quantizer."""
        super().__init__(quant_config)

    @staticmethod
    def update_quantizer_based_on_quant_config(quant_config=None) -> X86InductorQuantizer:
        """Updates the quantizer based on the given quantization configuration.

        Args:
            quant_config (dict): The quantization configuration. Defaults to None.

        Returns:
            X86InductorQuantizer: The updated quantizer object.
        """
        if not quant_config:
            quantizer = X86InductorQuantizer()
            quantizer.set_global(
                xiq.get_default_x86_inductor_quantization_config(is_dynamic=W8A8PT2EQuantizer.is_dynamic)
            )
        else:
            quantizer = create_xiq_quantizer_from_pt2e_config(quant_config, is_dynamic=W8A8PT2EQuantizer.is_dynamic)
        return quantizer

    def prepare(self, model: GraphModule, example_inputs=None, inplace=True, *args, **kwargs) -> GraphModule:
        """Prepares the model for calibration.

        Create the `quantizer` according to the `quant_config`, and insert the observers accordingly.

        Args:
            model (GraphModule): The model to be prepared for calibration.
            example_inputs (tuple, optional): Example inputs to be used for calibration. Defaults to None.
            inplace (bool, optional): Whether to modify the model in-place or return a new prepared model.
                Defaults to True.

        Returns:
            GraphModule: The prepared model.
        """
        quant_config = self.quant_config
        assert model._exported, "The model should be exported before preparing it for calibration."
        quantizer = self.update_quantizer_based_on_quant_config(quant_config)
        prepared_model = prepare_pt2e(model, quantizer)
        return prepared_model

    def convert(self, model: GraphModule, *args: Any, **kwargs: Any) -> GraphModule:
        """Convert the calibrated model into qdq mode.

        Args:
            model (GraphModule): The prepared model.

        Returns:
            GraphModule: The converted quantized model.
        """
        fold_quantize = kwargs.get("fold_quantize", False)
        converted_model = convert_pt2e(model, fold_quantize=fold_quantize)
        logger.warning("Converted the model in qdq mode, please compile it to accelerate inference.")
        if self.quant_config:
            self.half_precision_transformation(converted_model, self.quant_config)
        return converted_model

    def half_precision_transformation(self, model, config):
        """Applies half-precision transformation to the given model in-place.

        Args:
            model: The model to apply the transformation to.
            config: The configuration for the transformation.
        """
        half_precision_node_set = hp_rewriter.get_half_precision_node_set(model, config)
        logger.info("Try to convert %d nodes to half precision.", len(half_precision_node_set))
        hp_rewriter.transformation(model, half_precision_node_set, torch.float16)
        hp_rewriter.transformation(model, half_precision_node_set, torch.bfloat16)



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..quantized_func_wrapper import QuantizedFuncWrapperBase, OP_TYPE, QuantizedFuncWrapperFactory

import torch
from torchao.quantization.quant_primitives import (
    _quantize_affine_float8,
    _dequantize_affine_float8,
)

from abc import ABCMeta



class QuantizedCPUFuncWrapperBase(QuantizedFuncWrapperBase, metaclass=ABCMeta):
    """
    Placeholder for base class for CPU quantized func wrapper.
    """
    def __init__(self, scale_format, is_dynamic=False):
        self._quantized_func_ = self.get_default_quantized_func()


class QuantizedCPUQuant(QuantizedCPUFuncWrapperBase):

    def get_default_quantized_func(self):
        return _quantize_affine_float8

    def __call__(self, input, scale, zero_point=None, axis=0, quant_min=None, quant_max=None, dtype=torch.float8_e4m3fn):
        return self._quantized_func_(tensor=input, scale=torch.tensor(scale), float8_dtype=dtype)


class QuantizedCPUQuantPC(QuantizedCPUFuncWrapperBase):

    def get_default_quantized_func(self):
        return _quantize_affine_float8

    def __call__(self, input, scale, zero_point=None, axis=0, quant_min=None, quant_max=None, dtype=torch.float8_e4m3fn):
        return self._quantized_func_(tensor=input, scale=scale.view((-1, 1)), float8_dtype=dtype)


class QuantizedCPUDeQuant(QuantizedCPUFuncWrapperBase):

    def get_default_quantized_func(self):
        return _dequantize_affine_float8

    def __call__(self, input, scale, zero_point=None, axis=0, quant_min=None, quant_max=None, dtype=torch.float8_e4m3fn, out_dtype=torch.bfloat16):
        return self._quantized_func_(tensor=input, scale=torch.tensor(scale), output_dtype=out_dtype)


class QuantizedCPUDeQuantPC(QuantizedCPUFuncWrapperBase):

    def get_default_quantized_func(self):
        return _dequantize_affine_float8

    def __call__(self, input, scale, zero_point=None, axis=0, quant_min=None, quant_max=None, dtype=torch.float8_e4m3fn, out_dtype=torch.bfloat16):
        return self._quantized_func_(tensor=input, scale=scale.view((1, -1)), output_dtype=out_dtype)


_OP_TYPE_CPU_QUANTIZED_WRAPPER_CLASSES = {
                                          OP_TYPE.QUANT: QuantizedCPUQuant,
                                          OP_TYPE.DEQUANT: QuantizedCPUDeQuant,
                                          OP_TYPE.QUANT_PC: QuantizedCPUQuantPC,
                                          OP_TYPE.DEQUANT_PC: QuantizedCPUDeQuantPC,
                                         }


def init_cpu_quantized_func_wrapper_factory():
    QuantizedFuncWrapperFactory.initialize(_OP_TYPE_CPU_QUANTIZED_WRAPPER_CLASSES)



"""Public entry points for the FP8 quantization algorithm."""

# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neural_compressor.torch.algorithms.fp8_quant.common import (
    update_mode,
    save_calib_result,
    restore_patched_module,
    with_patched_module,
)
from neural_compressor.torch.algorithms.fp8_quant.prepare_quant.prepare_model import finish_measurements, prep_model
from neural_compressor.torch.algorithms.fp8_quant.quantizer import FP8Quantizer
from neural_compressor.torch.algorithms.fp8_quant.patched_module_base import (
    PatchedModuleBase,
    register_patched_module,
)

from neural_compressor.torch.algorithms.fp8_quant.scaling_method_base import (
    ScalingMethodBase,
    register_scaling_methods,
)
from neural_compressor.torch.algorithms.fp8_quant.observer import (
    ObserverBase,
    register_observer,
)
from neural_compressor.torch.algorithms.fp8_quant.model_configs import (
    ModuleConfig,
    ModuleInfo,
    ModuleType,
    ModuleExtraConfig
)
from neural_compressor.torch.algorithms.fp8_quant.save_load import save, load



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
#
# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized Linear."""


import torch
import torch.nn as nn
import torch.nn.functional as F

from .tensor_quantizer import TensorQuantizer


class QuantLinear(nn.Module):
    """Quantized version of nn.Linear."""

    def forward(self, input: torch.Tensor):
        """Add weight/input/output of quantization for the original forward method."""
        qw = self.weight_quantizer(self.weight)
        qi = self.input_quantizer(input)
        out = F.linear(qi, qw, self.bias)
        out = self.output_quantizer(out)
        return out

    def _setup(self, quant_cfg):
        """Init quantizer."""
        self.weight_quantizer = TensorQuantizer(
            data_type=quant_cfg.data_type,
            block_size=quant_cfg.group_size,
            bits=quant_cfg.bits,
            sym=quant_cfg.sym,
            if_quant=True,
            learn_exponent=False,
        )
        self.input_quantizer = TensorQuantizer(
            data_type=quant_cfg.act_data_type,
            block_size=quant_cfg.act_group_size,
            bits=quant_cfg.act_bits,
            sym=quant_cfg.act_sym,
            if_quant=True,
            learn_exponent=False,
        )
        self.output_quantizer = TensorQuantizer(
            data_type=quant_cfg.act_data_type,
            block_size=quant_cfg.act_group_size,
            bits=quant_cfg.act_bits,
            sym=quant_cfg.act_sym,
            if_quant=False,
        )
        # Currently don't quant output
        self.output_quantizer.disable()

        # TODO: remove
        self.original_weight_dtype = None if self.weight is None else self.weight.dtype

    def extra_repr(self) -> str:
        """Generate extra_repr making sure import keys exist in self.__dict__."""
        return f"in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}"

    def __repr__(self):
        """Overriding the __repr__ method, makes the output more concise and meaningful."""
        return (
            f"QuantLinear(\n"
            f"  in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}\n"
            f"  (input_quantizer): {self.input_quantizer}\n"
            f"  (output_quantizer): {self.output_quantizer}\n"
            f"  (weight_quantizer): {self.weight_quantizer}\n"
            f")"
        )



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neural_compressor.torch.algorithms.fp8_quant.utils.logger import logger
from neural_compressor.torch.algorithms.fp8_quant._quant_common.quant_config import get_hqt_config, QuantMode
from neural_compressor.torch.algorithms.fp8_quant._core.common import save_file, save_json


def is_measure_done(mod_extra_config):
    # check if measurements were collected by observer
    for obs in ([] if mod_extra_config.inputs is None else mod_extra_config.inputs) + (
        [] if mod_extra_config.outputs is None else mod_extra_config.outputs
    ):
        if obs.is_used():
            return True
    return False

def get_mod_extra_config_dict(model):
    mcd = {}
    for name, mod in model.named_modules():
        if hasattr(mod, "_mod_extra_config") and mod._mod_extra_config:
            if is_measure_done(mod._mod_extra_config):
                name = name.replace("_orig_mod.", "")  # remove _orig_mod part added by dynamo mechanism
                mcd[name] = mod._mod_extra_config
            else:
                logger.debug(
                    "Layer '%s' has no measurements therefore it can't be quantized during quantization.",
                    name,
                )
    return mcd

def measure_control_to_state_dict(mcd):
    sd = {}
    sdl = {}
    for mname in mcd:
        sd[mname] = dict()
        sdl[mname] = dict()
        sd[mname]["inputs"] = [
            mcd[mname].inputs[i].state.detach().cpu().float().numpy()
            for i in range(len(mcd[mname].inputs))
            if mcd[mname].inputs[i].state is not None
        ]
        sdl[mname]["inputs"] = [
            mcd[mname].inputs[i].state.detach().cpu().float().numpy().tolist()
            for i in range(len(mcd[mname].inputs))
            if mcd[mname].inputs[i].state is not None
        ]
        if mcd[mname].outputs:
            sd[mname]["outputs"] = [
                mcd[mname].outputs[i].state.detach().cpu().float().numpy()
                for i in range(len(mcd[mname].outputs))
                if mcd[mname].outputs[i].state is not None
            ]
            sdl[mname]["outputs"] = [
                mcd[mname].outputs[i].state.detach().cpu().float().numpy().tolist()
                for i in range(len(mcd[mname].outputs))
                if mcd[mname].outputs[i].state is not None
            ]
        if len(mcd[mname].params) > 0:
            sd[mname]["params"] = dict()
            sdl[mname]["params"] = dict()
            for param_name in mcd[mname].params:
                if mcd[mname].params[param_name].state is not None:
                    sd[mname]["params"][param_name] = mcd[mname].params[param_name].state.detach().cpu().float().numpy()
                    sdl[mname]["params"][param_name] = (
                        mcd[mname].params[param_name].state.detach().cpu().float().numpy().tolist()
                    )
    return sd, sdl

def create_files_names(config, fname = None):
    if fname is None:
        if ("measure_file" in config) and (config["measure_file"] is not None):
            fname_base = config["measure_file"]
            measure_type = "DynamicRange"
        elif ("shape_file" in config) and (config["shape_file"] is not None) and (config["observer"] == "shape"):
            fname_base = config["shape_file"]
            measure_type = "Shape"
        fname_np = fname_base + ".npz"
        fname_list = fname_base + ".json"
        return fname_base, fname_np, fname_list, measure_type
    else:
        logger.warning("'fname' is not None - Measurements/Shapes will not be saved")
        return

def save_measurements_files(model, state_dict, state_list, gmod_list, fname_np, fname_list, fname_base, measure_type,
                            num_samples=0):
    import numpy as np
    logger.info("Dumping measurements")
    save_file(model, state_dict, np.ndarray, fname_np, measure_type, num_samples)
    save_file(model, state_list, list, fname_list, measure_type, num_samples)
    save_json(gmod_list, fname_base + "_mod_list.json")
    return


gmod_list = [] # global list extened with patched modules in measure.prepare_model


def save_measurements(model, fname=None):
    config = get_hqt_config(model).cfg
    if config["mode"] in [QuantMode.MEASURE, QuantMode.SHAPE]:
        fname_base, fname_np, fname_list, measure_type = create_files_names(config, fname)
        mcd = get_mod_extra_config_dict(model)
        sd, sdl = measure_control_to_state_dict(mcd)
        num_samples = model.calibration_samples_counter if hasattr(model, "calibration_samples_counter") else 0
        save_measurements_files(model, sd, sdl, gmod_list, fname_np, fname_list, fname_base, measure_type, num_samples)



#
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Half-precision Wrapper for Torch Modules."""

import torch


class HalfPrecisionModuleWrapper(torch.nn.Module):
    """FP16 or BF16 Module Wrapper Class."""

    def __init__(self, module, device="cpu", dtype=torch.float16):
        """Init a HalfPrecisionModuleWrapper object."""
        super(HalfPrecisionModuleWrapper, self).__init__()
        self.add_module("module", module)
        self.device = device
        self.dtype = dtype
        self.weight = self.module.weight if hasattr(self.module, "weight") else None
        self.bias = self.module.bias if hasattr(self.module, "bias") else None

    def forward(self, X):
        """Convert dtype."""
        with torch.autocast(device_type=self.device, dtype=self.dtype):
            X = self.module(X)
        return X.float()



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intel Neural Compressor PyTorch Quantizer API."""

import copy
from abc import ABC, abstractmethod
from typing import Any, Optional

import torch

from neural_compressor.common.utils import Mode


class Quantizer(ABC):
    """The base quantizer for all algorithm quantizers.

    The `Quantizer` unifies the interfaces across various quantization algorithms, including GPTQ, RTN, etc.
    Given a float model, `Quantizer` apply the quantization algorithm to the model according to the `quant_config`.

    To implement a new quantization algorithm,, inherit from `Quantizer` and implement the following methods:
        - `prepare`: prepare a given model for convert.
        - `convert`: convert a prepared model to a quantized model.
    Note: `quantize` and `execute` are optional for new quantization algorithms.
    """

    def __init__(self, quant_config: Optional[Any] = None):
        """Init a Quantizer object.

        Args:
            quant_config : Specifies how to apply the algorithm on the given model.
            The format of `quant_config` can be defined by `Quantized` itself.
            For example, `quant_config` can be a dictionary as below:
                quant_config={
                'fc2':{
                    'dtype': 'int',
                    'bits': 4,
                    'group_size': 32,
                    'scheme': 'sym'
                    }}
        """
        self.quant_config = quant_config

    @abstractmethod
    def prepare(self, model: torch.nn.Module, *args: Any, **kwargs: Any):
        """Prepares a given model for quantization.

        Insert observers into the model so that it can monitor the input and output tensors during calibration.

        Args:
            model (torch.nn.Module): The model to be prepared.

        Returns:
            A prepared model.
        """
        raise NotImplementedError("{} doesn't implement `prepare` function. ".format(self.__class__.__name__))

    @abstractmethod
    def convert(self, model: torch.nn.Module, *args: Any, **kwargs: Any):
        """Converts a prepared model to a quantized model.

        Args:
            model (torch.nn.Module): The prepared model to be converted.

        Returns:
            A quantized model.
        """
        raise NotImplementedError("{} doesn't implement `convert` function. ".format(self.__class__.__name__))

    def quantize(self, model: torch.nn.Module, *args: Any, **kwargs: Any):
        """Quantizes a given float model.

        Args:
            model (torch.nn.Module): The float model to be quantized.

        Returns:
            A quantized model.
        """
        model = self.prepare(model, *args, **kwargs)

        run_fn = kwargs.get("run_fn", None)
        if run_fn is not None:
            run_args = kwargs.get("run_args", None)
            if run_args:
                run_fn(model, *run_args)
            else:
                run_fn(model)

        model = self.convert(model, *args, **kwargs)

        return model

    def execute(self, model: torch.nn.Module, mode, *args: Any, **kwargs: Any):
        """Execute according to mode.

        Args:
            model (torch.nn.Module): The model to be executed.
            mode (Mode): The mode of current phase, including 'prepare', 'convert' and 'quantize'.
        """
        if mode == Mode.PREPARE:
            model = self.prepare(model, *args, **kwargs)
        elif mode == Mode.CONVERT:
            model = self.convert(model, *args, **kwargs)
        elif mode == Mode.QUANTIZE:
            if not isinstance(self.quant_config, dict):
                user_cfg = copy.deepcopy(self.quant_config).to_dict()
            else:
                user_cfg = copy.deepcopy(self.quant_config)
            if "recipe_cfgs" in user_cfg:  # keep quantize API for smoothquant
                run_fn = kwargs.get("run_fn", None)
                example_inputs = kwargs.get("example_inputs", None)
                inplace = kwargs.get("inplace", True)
                model = self.quantize(model, self.quant_config, run_fn, example_inputs, inplace)
            else:
                model = self.quantize(model, *args, **kwargs)
        return model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""HQQ Quantizer."""


from typing import Callable, List, Optional, Tuple

import torch

from neural_compressor.torch.algorithms import Quantizer
from neural_compressor.torch.utils import logger
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

from .config import ConfigMappingType, HQQModuleConfig, QTensorConfig, hqq_global_option
from .core import HQQLinear


def _has_child(module: torch.nn.Module) -> bool:
    return len(list(module.named_children())) > 0


def _replace_with_custom_fn_if_matches_filter(
    model: torch.nn.Module,
    replacement_fn: Callable,
    filter_fn: Callable,
    cur_fqn: str = "",
    config_mapping: Optional[ConfigMappingType] = None,
) -> None:
    """Recursively replaces modules in `model` with `replacement_fn` if `filter_fn` is `True`."""
    name_to_child = dict(model.named_children())
    for name, child in name_to_child.items():
        if cur_fqn == "":
            new_fqn = name
        else:
            new_fqn = f"{cur_fqn}.{name}"
        if filter_fn(child, new_fqn, config_mapping):
            new_child = replacement_fn(child.to(auto_detect_accelerator().current_device()), new_fqn, config_mapping)
            logger.debug("Quantize linear module %s.", new_fqn)
            setattr(model, name, new_child)
        elif not _has_child(child):  # TODO: merge it into `filter_fn`
            if hqq_global_option.use_half:
                logger.debug("Half module %s.", new_fqn)
                child = child.half()
            new_child = child.to(auto_detect_accelerator().current_device())
            setattr(model, name, new_child)
        else:
            _replace_with_custom_fn_if_matches_filter(
                model=child,
                replacement_fn=replacement_fn,
                filter_fn=filter_fn,
                cur_fqn=new_fqn,
                config_mapping=config_mapping,
            )


def patch_hqq_moduile(mod, config):
    """Patch the given module with the HQQLinear module.

    Args:
        mod (torch.nn.Module): The module to be patched.
        config (dict): Configuration parameters for the HQQLinear module.

    Returns:
        torch.nn.Module: The patched module with HQQLinear.
    """
    new_mod = HQQLinear.from_float(mod, config)
    return new_mod


def filter_fn(mod: torch.nn.Module, name: str, config_mapping: ConfigMappingType) -> bool:
    """Filter function used to determine if a module should be quantized.

    Args:
        mod (torch.nn.Module): The module to be checked.
        name (str): The name of the module.
        config_mapping (ConfigMappingType): The configuration mapping.

    Returns:
        bool: True if the module should be quantized, False otherwise.
    """
    return isinstance(mod, torch.nn.Linear) and name in config_mapping


def replacement_fn(mod: torch.nn.Module, name: str, config_mapping: ConfigMappingType) -> torch.nn.Module:
    """Replaces a Linear with HQQLinear if the module is in the config mapping.

    Args:
        mod (torch.nn.Module): The original module to be replaced.
        name (str): The name of the module to be replaced.
        config_mapping (ConfigMappingType): A mapping of module names to their corresponding configurations.

    Returns:
        torch.nn.Module: The patched module.
    """
    config = config_mapping.get(name, None)
    logger.debug("Replace module %s", name)
    return patch_hqq_moduile(mod, config)


class HQQuantizer(Quantizer):
    """HQQ Quantizer."""

    def __init__(self, quant_config: ConfigMappingType) -> None:
        """Init a HQQuantizer object.

        Args:
            quant_config (ConfigMappingType): quantization config for ops.
        """
        quant_config = self._parse_hqq_configs_mapping(quant_config)
        super().__init__(quant_config=quant_config)

    @torch.no_grad()
    def prepare(self, model: torch.nn.Module, *args, **kwargs) -> Optional[torch.nn.Module]:
        """Prepares a given model for quantization.

        Will return model directly in HQQ algorithm.

        Args:
            model (torch.nn.Module): The model to be prepared.
        """
        return model

    @torch.no_grad()
    def convert(self, model: torch.nn.Module, *args, **kwargs) -> Optional[torch.nn.Module]:
        """Converts a prepared model to a quantized model.

        Args:
            model (torch.nn.Module): The prepared model to be converted.

        Returns:
            Optional[torch.nn.Module]: A quantized model.
        """
        _replace_with_custom_fn_if_matches_filter(
            model, replacement_fn=replacement_fn, filter_fn=filter_fn, config_mapping=self.quant_config
        )
        return model

    def _convert_hqq_module_config(self, config) -> HQQModuleConfig:
        # TODO: (Yi) Please note that the configuration defined by INC should be separated from the algorithm.
        # * 3.x API use `bits` for woq while HQQ internal API use `nbits`, we should change it in algorithm_entry.py
        nbits = config.bits
        group_size = config.group_size
        quant_zero = config.quant_zero
        quant_scale = config.quant_scale
        scale_quant_group_size = config.scale_quant_group_size

        weight_qconfig = QTensorConfig(
            nbits=nbits,
            channel_wise=True,
            group_size=group_size,
            optimize=True,
            round_zero=True if nbits == 4 else False,
        )
        zero_qconfig = None
        if quant_zero:
            zero_qconfig = QTensorConfig(nbits=8, channel_wise=False, group_size=None, optimize=False)
        scale_qconfig = None
        if quant_scale:
            scale_qconfig = QTensorConfig(nbits=8, channel_wise=True, group_size=scale_quant_group_size, optimize=False)
        hqq_module_config = HQQModuleConfig(weight=weight_qconfig, scale=scale_qconfig, zero=zero_qconfig)
        logger.debug(hqq_module_config)
        return hqq_module_config

    def _parse_hqq_configs_mapping(self, configs_mapping):
        qconfig_mapping = {}
        for (op_name, op_type), quant_config in configs_mapping.items():
            if quant_config is not None and quant_config.dtype == "fp32":
                logger.warning("Fallback %s.", op_name)
                continue
            qconfig_mapping[op_name] = self._convert_hqq_module_config(quant_config)
        return qconfig_mapping



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The quantizer using StaticQuant path."""


import json
import os
from copy import deepcopy
from types import MethodType

import torch

try:
    import intel_extension_for_pytorch as ipex
except:  # pragma: no cover
    assert False, "Please install IPEX for static quantization."

from collections import OrderedDict

from packaging.version import Version

from neural_compressor.torch.algorithms import Quantizer
from neural_compressor.torch.utils import logger
from neural_compressor.torch.utils.auto_accelerator import auto_detect_accelerator

from .utility import (
    CpuInfo,
    cfg_to_qconfig,
    dump_model_op_stats,
    generate_xpu_qconfig,
    get_ipex_version,
    get_quantizable_ops_recursively,
    ipex_config_path,
    simple_inference,
)

ipex_ver = get_ipex_version()


class StaticQuantQuantizer(Quantizer):
    """The quantizer using Static Quant."""

    def __init__(self, quant_config: OrderedDict = {}):
        """Init a StaticQuantQuantizer object.

        Args:
            quant_config (OrderedDict, optional): quantization config for ops. Defaults to {}.
        """
        super().__init__(quant_config)
        self.user_cfg = OrderedDict()
        self.device = auto_detect_accelerator().current_device()

    def prepare(self, model, example_inputs, inplace=True, *args, **kwargs):
        """Prepares a given model for quantization.

        Args:
            model (torch.nn.Module): raw fp32 model or prepared model.
            example_inputs (tensor/tuple/dict): used to trace torch model.
            inplace (bool, optional): whether to carry out model transformations in-place. Defaults to True.

        Returns:
            A prepared model.
        """
        assert example_inputs is not None, "Please provide example_inputs for static quantization."

        if self.device == "cpu":
            _, cfgs, op_infos_from_cfgs, output_tensor_id_op_name, _ = get_quantizable_ops_recursively(
                model, example_inputs
            )
            # update json file in ipex_config_path; map ipex op_name to pt op_name
            self.user_cfg = cfg_to_qconfig(self.quant_config, cfgs, op_infos_from_cfgs, output_tensor_id_op_name)
        else:  # pragma: no cover
            model = model.to("xpu")

        model.eval()

        # Check save_qconf_summary part is a workaround for IPEX bug.
        # Sometimes the prepared model from get_op_capablitiy loss this attributes
        if not hasattr(model, "save_qconf_summary") or not hasattr(model, "load_qconf_summary"):  # pragma: no cover
            from torch.ao.quantization import HistogramObserver, MinMaxObserver, PerChannelMinMaxObserver, QConfig

            if self.device != "cpu":  # pragma: no cover
                from torch.quantization.quantize_jit import prepare_jit

                with torch.no_grad():
                    modelJit = torch.jit.trace(model, example_inputs)
                qconfig = generate_xpu_qconfig(self.quant_config)
                model = prepare_jit(modelJit, qconfig, inplace)
            else:
                if ipex_ver.release >= Version("2.1").release:
                    # HistogramObserver will cause a performance issue.
                    # static_qconfig = ipex.quantization.default_static_qconfig_mapping
                    qconfig = QConfig(
                        activation=MinMaxObserver.with_args(qscheme=torch.per_tensor_affine, dtype=torch.quint8),
                        weight=PerChannelMinMaxObserver.with_args(
                            dtype=torch.qint8, qscheme=torch.per_channel_symmetric
                        ),
                    )
                    from torch.ao.quantization import QConfigMapping

                    static_qconfig = QConfigMapping().set_global(qconfig)
                else:  # pragma: no cover
                    static_qconfig = QConfig(
                        activation=MinMaxObserver.with_args(qscheme=torch.per_tensor_affine, dtype=torch.quint8),
                        weight=PerChannelMinMaxObserver.with_args(
                            dtype=torch.qint8, qscheme=torch.per_channel_symmetric
                        ),
                    )
                if isinstance(example_inputs, dict):
                    model = ipex.quantization.prepare(
                        model, static_qconfig, example_kwarg_inputs=example_inputs, inplace=inplace
                    )
                else:
                    model = ipex.quantization.prepare(
                        model, static_qconfig, example_inputs=example_inputs, inplace=inplace
                    )

        if self.device == "cpu":
            model.load_qconf_summary(qconf_summary=ipex_config_path)

        return model

    def convert(self, model, example_inputs, inplace=True, *args, **kwargs):
        """Converts a prepared model to a quantized model.

        Args:
            model (QuantizationInterceptionModule): the prepared model to be converted.
            example_inputs (tensor/tuple/dict): used to trace torch model.
            inplace (bool, optional): whether to carry out model transformations in-place. Defaults to True.

        Returns:
            A quantized model.
        """
        use_bf16 = self.quant_config.get("use_bf16", None)

        from neural_compressor.torch.algorithms.static_quant import save

        if self.device != "cpu":  # pragma: no cover
            from torch.quantization.quantize_jit import convert_jit

            model = convert_jit(model, inplace)
            simple_inference(model, example_inputs, iterations=2)
            model.qconfig = self.quant_config["op"]
            dump_model_op_stats(model.qconfig)
        else:
            model.save_qconf_summary(qconf_summary=ipex_config_path)
            model = _ipex_post_quant_process(model, example_inputs, use_bf16, inplace=inplace)

            with open(ipex_config_path, "r") as f:
                model.tune_cfg = json.load(f)
            model.ipex_config_path = ipex_config_path

            dump_model_op_stats(self.user_cfg)

        model.ori_save = model.save
        model.save = MethodType(save, model)

        logger.info("Static quantization done.")
        return model


def _ipex_post_quant_process(model, example_inputs, use_bf16, inplace=False):
    """Converts to a jit model.

    Args:
        model (QuantizationInterceptionModule): a prepared model.
        example_inputs (tensor/tuple/dict): used to trace torch model.
        use_bf16 (bool): whether to use bf16 for mixed precision.
        inplace (bool, optional): whether to carry out model transformations in-place. Defaults to True.

    Returns:
        A converted jit model.
    """
    if (
        use_bf16 and (not os.getenv("FORCE_FP32") == "1") and (CpuInfo().bf16 or os.getenv("FORCE_BF16") == "1")
    ):  # pragma: no cover
        with torch.no_grad():
            with torch.cpu.amp.autocast():
                model = ipex.quantization.convert(model, inplace=inplace)
                try:
                    if isinstance(example_inputs, dict):
                        model = torch.jit.trace(model, example_kwarg_inputs=example_inputs)
                    else:
                        model = torch.jit.trace(model, example_inputs)
                    model = torch.jit.freeze(model.eval())
                except:
                    if isinstance(example_inputs, dict):
                        model = torch.jit.trace(
                            model, example_kwarg_inputs=example_inputs, strict=False, check_trace=False
                        )
                    else:
                        model = torch.jit.trace(model, example_inputs, strict=False)
                    model = torch.jit.freeze(model.eval())
    else:  # pragma: no cover
        model = ipex.quantization.convert(model, inplace=inplace)
        with torch.no_grad():
            try:
                if isinstance(example_inputs, dict):
                    model = torch.jit.trace(model, example_kwarg_inputs=example_inputs)
                else:
                    model = torch.jit.trace(model, example_inputs)
                model = torch.jit.freeze(model.eval())
            except:
                if isinstance(example_inputs, dict):
                    model = torch.jit.trace(model, example_kwarg_inputs=example_inputs, strict=False, check_trace=False)
                else:
                    model = torch.jit.trace(model, example_inputs, strict=False)
                model = torch.jit.freeze(model.eval())

    # After freezing, run 1 time to warm up the profiling graph executor to insert prim::profile
    # At the 2nd run, the llga pass will be triggered and the model is turned into
    # an int8 model: prim::profile will be removed and will have LlgaFusionGroup in the graph
    simple_inference(model, example_inputs, iterations=2)
    return model



# Copyright (c) 2024 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
import time
import warnings
from typing import List, Optional, Tuple, Union

import torch
import torch.distributed as dist
from torch import nn
from transformers.generation.beam_search import BeamScorer
from transformers.generation.logits_process import LogitsProcessorList
from transformers.generation.stopping_criteria import StoppingCriteriaList, validate_stopping_criteria
from transformers.utils import ModelOutput


class BeamSearchEncoderDecoderOutput(ModelOutput):
    sequences: torch.LongTensor = None
    sequences_scores: Optional[torch.FloatTensor] = None
    scores: Optional[Tuple[torch.FloatTensor]] = None
    beam_indices: Optional[torch.LongTensor] = None
    encoder_attentions: Optional[Tuple[torch.FloatTensor]] = None
    encoder_hidden_states: Optional[Tuple[torch.FloatTensor]] = None
    decoder_attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    cross_attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    decoder_hidden_states: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    past_key_values: Optional[Tuple[Tuple[Tuple[torch.FloatTensor]]]] = None


class BeamSearchDecoderOnlyOutput(ModelOutput):
    sequences: torch.LongTensor = None
    sequences_scores: Optional[torch.FloatTensor] = None
    scores: Optional[Tuple[torch.FloatTensor]] = None
    beam_indices: Optional[torch.LongTensor] = None
    attentions: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    hidden_states: Optional[Tuple[Tuple[torch.FloatTensor]]] = None
    past_key_values: Optional[Tuple[Tuple[Tuple[torch.FloatTensor]]]] = None


BeamSearchOutput = Union[BeamSearchEncoderDecoderOutput, BeamSearchDecoderOnlyOutput]


def _beam_search(
    self,
    input_ids: torch.LongTensor,
    beam_scorer: BeamScorer,
    logits_processor: Optional[LogitsProcessorList] = None,
    stopping_criteria: Optional[StoppingCriteriaList] = None,
    max_length: Optional[int] = None,
    pad_token_id: Optional[int] = None,
    eos_token_id: Optional[Union[int, List[int]]] = None,
    output_attentions: Optional[bool] = None,
    output_hidden_states: Optional[bool] = None,
    output_scores: Optional[bool] = None,
    return_dict_in_generate: Optional[bool] = None,
    synced_gpus: bool = False,
    **model_kwargs,
) -> Union[BeamSearchOutput, torch.LongTensor]:
    r"""
    Generates sequences of token ids for models with a language modeling head using **beam search decoding** and
    can be used for text-decoder, text-to-text, speech-to-text, and vision-to-text models.
    <Tip warning={true}>
    In most cases, you do not need to call [`~generation.GenerationMixin.beam_search`] directly. Use generate()
    instead. For an overview of generation strategies and code examples, check the [following
    guide](../generation_strategies).
    </Tip>
    Parameters:
        input_ids (`torch.LongTensor` of shape `(batch_size, sequence_length)`):
            The sequence used as a prompt for the generation.
        beam_scorer (`BeamScorer`):
            An derived instance of [`BeamScorer`] that defines how beam hypotheses are constructed, stored and
            sorted during generation. For more information, the documentation of [`BeamScorer`] should be read.
        logits_processor (`LogitsProcessorList`, *optional*):
            An instance of [`LogitsProcessorList`]. List of instances of class derived from [`LogitsProcessor`]
            used to modify the prediction scores of the language modeling head applied at each generation step.
        stopping_criteria (`StoppingCriteriaList`, *optional*):
            An instance of [`StoppingCriteriaList`]. List of instances of class derived from [`StoppingCriteria`]
            used to tell if the generation loop should stop.
        max_length (`int`, *optional*, defaults to 20):
            **DEPRECATED**. Use `logits_processor` or `stopping_criteria` directly to cap the number of generated
            tokens. The maximum length of the sequence to be generated.
        pad_token_id (`int`, *optional*):
            The id of the *padding* token.
        eos_token_id (`Union[int, List[int]]`, *optional*):
            The id of the *end-of-sequence* token. Optionally, use a list to set multiple *end-of-sequence* tokens.
        output_attentions (`bool`, *optional*, defaults to `False`):
            Whether or not to return the attentions tensors of all attention layers. See `attentions` under
            returned tensors for more details.
        output_hidden_states (`bool`, *optional*, defaults to `False`):
            Whether or not to return the hidden states of all layers. See `hidden_states` under returned tensors
            for more details.
        output_scores (`bool`, *optional*, defaults to `False`):
            Whether or not to return the prediction scores. See `scores` under returned tensors for more details.
        return_dict_in_generate (`bool`, *optional*, defaults to `False`):
            Whether or not to return a [`~utils.ModelOutput`] instead of a plain tuple.
        synced_gpus (`bool`, *optional*, defaults to `False`):
            Whether to continue running the while loop until max_length (needed for ZeRO stage 3)
        model_kwargs:
            Additional model specific kwargs will be forwarded to the `forward` function of the model. If model is
            an encoder-decoder model the kwargs should include `encoder_outputs`.
    Return:
        [`BeamSearchDecoderOnlyOutput`], [`BeamSearchEncoderDecoderOutput`] or
        `torch.LongTensor`: A `torch.LongTensor` containing the generated tokens (default behaviour) or a
        [`BeamSearchDecoderOnlyOutput`] if `model.config.is_encoder_decoder=False` and
        `return_dict_in_generate=True` or a [`BeamSearchEncoderDecoderOutput`] if
        `model.config.is_encoder_decoder=True`.
    Examples:
    ```python
    >>> from transformers import (
    ...     AutoTokenizer,
    ...     AutoModelForSeq2SeqLM,
    ...     LogitsProcessorList,
    ...     MinLengthLogitsProcessor,
    ...     BeamSearchScorer,
    ... )
    >>> import torch
    >>> tokenizer = AutoTokenizer.from_pretrained("t5-base")
    >>> model = AutoModelForSeq2SeqLM.from_pretrained("t5-base")
    >>> encoder_input_str = "translate English to German: How old are you?"
    >>> encoder_input_ids = tokenizer(encoder_input_str, return_tensors="pt").input_ids
    >>> # lets run beam search using 3 beams
    >>> num_beams = 3
    >>> # define decoder start token ids
    >>> input_ids = torch.ones((num_beams, 1), device=model.device, dtype=torch.long)
    >>> input_ids = input_ids * model.config.decoder_start_token_id
    >>> # add encoder_outputs to model keyword arguments
    >>> model_kwargs = {
    ...     "encoder_outputs": model.get_encoder()(
    ...         encoder_input_ids.repeat_interleave(num_beams, dim=0), return_dict=True
    ...     )
    ... }
    >>> # instantiate beam scorer
    >>> beam_scorer = BeamSearchScorer(
    ...     batch_size=1,
    ...     num_beams=num_beams,
    ...     device=model.device,
    ... )
    >>> # instantiate logits processors
    >>> logits_processor = LogitsProcessorList(
    ...     [
    ...         MinLengthLogitsProcessor(5, eos_token_id=model.config.eos_token_id),
    ...     ]
    ... )
    >>> outputs = model.beam_search(input_ids, beam_scorer, logits_processor=logits_processor, **model_kwargs)
    >>> tokenizer.batch_decode(outputs, skip_special_tokens=True)
    ['Wie alt bist du?']
    ```"""
    # init values
    token_latency = (self.config.token_latency if hasattr(self.config, "token_latency") else False) or (
        self.token_latency if hasattr(self, "token_latency") else False
    )

    latency_list = []
    logits_processor = logits_processor if logits_processor is not None else LogitsProcessorList()
    stopping_criteria = stopping_criteria if stopping_criteria is not None else StoppingCriteriaList()
    if max_length is not None:
        warnings.warn(
            "`max_length` is deprecated in this function, use"
            " `stopping_criteria=StoppingCriteriaList([MaxLengthCriteria(max_length=max_length)])` instead.",
            UserWarning,
        )
        stopping_criteria = validate_stopping_criteria(stopping_criteria, max_length)
    if len(stopping_criteria) == 0:
        warnings.warn("You don't have defined any stopping_criteria, this will likely loop forever", UserWarning)
    pad_token_id = pad_token_id if pad_token_id is not None else self.generation_config.pad_token_id
    eos_token_id = eos_token_id if eos_token_id is not None else self.generation_config.eos_token_id
    if isinstance(eos_token_id, int):
        eos_token_id = [eos_token_id]
    output_scores = output_scores if output_scores is not None else self.generation_config.output_scores
    output_attentions = output_attentions if output_attentions is not None else self.generation_config.output_attentions
    output_hidden_states = (
        output_hidden_states if output_hidden_states is not None else self.generation_config.output_hidden_states
    )
    return_dict_in_generate = (
        return_dict_in_generate
        if return_dict_in_generate is not None
        else self.generation_config.return_dict_in_generate
    )

    batch_size = len(beam_scorer._beam_hyps)
    num_beams = beam_scorer.num_beams

    batch_beam_size, cur_len = input_ids.shape

    if num_beams * batch_size != batch_beam_size:
        raise ValueError(
            f"Batch dimension of `input_ids` should be {num_beams * batch_size}, but is {batch_beam_size}."
        )

    # init attention / hidden states / scores tuples
    scores = () if (return_dict_in_generate and output_scores) else None
    beam_indices = tuple(() for _ in range(batch_beam_size)) if (return_dict_in_generate and output_scores) else None
    decoder_attentions = () if (return_dict_in_generate and output_attentions) else None
    cross_attentions = () if (return_dict_in_generate and output_attentions) else None
    decoder_hidden_states = () if (return_dict_in_generate and output_hidden_states) else None
    # if model is an encoder-decoder, retrieve encoder attention weights and hidden states
    if return_dict_in_generate and self.config.is_encoder_decoder:
        encoder_attentions = model_kwargs["encoder_outputs"].get("attentions") if output_attentions else None
        encoder_hidden_states = model_kwargs["encoder_outputs"].get("hidden_states") if output_hidden_states else None
    # initialise score of first beam with 0 and the rest with -1e9. This makes sure that only tokens
    # of the first beam are considered to avoid sampling the exact same tokens across all beams.
    beam_scores = torch.zeros((batch_size, num_beams), dtype=torch.float, device=input_ids.device)
    beam_scores[:, 1:] = -1e9
    beam_scores = beam_scores.view((batch_size * num_beams,))
    this_peer_finished = False  # used by synced_gpus only
    decoder_prompt_len = input_ids.shape[-1]  # record the prompt length of decoder
    while True:
        tic = time.time()
        if synced_gpus:
            # Under synced_gpus the `forward` call must continue until all gpus complete their sequence.
            # The following logic allows an early break if all peers finished generating their sequence
            this_peer_finished_flag = torch.tensor(0.0 if this_peer_finished else 1.0).to(input_ids.device)
            # send 0.0 if we finished, 1.0 otherwise
            dist.all_reduce(this_peer_finished_flag, op=dist.ReduceOp.SUM)
            # did all peers finish? the reduced sum will be 0.0 then
            if this_peer_finished_flag.item() == 0.0:
                break

        model_inputs = self.prepare_inputs_for_generation(input_ids, **model_kwargs)
        if (
            re.search("GPTJ", self.config.architectures[0])
            or re.search("llama", self.config.architectures[0], re.IGNORECASE)
            or re.search("gptneox", self.config.architectures[0], re.IGNORECASE)
            or re.search("OPT", self.config.architectures[0], re.IGNORECASE)
            or re.search("falcon", self.config.architectures[0], re.IGNORECASE)
            or re.search("rw", self.config.architectures[0], re.IGNORECASE)
        ):
            first_token = False
            input_bs = input_ids.size()[0]
            has_position_id = True
            if model_inputs["past_key_values"] is None:
                first_token = True
            if first_token and hasattr(self, "trace_graph"):
                if re.search("GPTJ", self.config.architectures[0]):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(batch_size * num_beams)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.n_layer)
                        ]
                    )
                elif re.search("llama", self.config.architectures[0], re.IGNORECASE):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(batch_size * num_beams)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.num_hidden_layers)
                        ]
                    )
                elif re.search("gptneox", self.config.architectures[0], re.IGNORECASE):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(batch_size * num_beams)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.num_hidden_layers)
                        ]
                    )
                elif re.search("OPT", self.config.architectures[0], re.IGNORECASE):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(batch_size * num_beams)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.num_hidden_layers)
                        ]
                    )
                    has_position_id = False
                elif re.search("falcon", self.config.architectures[0], re.IGNORECASE) or re.search(
                    "rw", self.config.architectures[0], re.IGNORECASE
                ):
                    beam_idx_tmp = torch.zeros(
                        (2048, int(batch_size * num_beams)), dtype=torch.long, device=input_ids.device
                    ).contiguous()
                    model_inputs["past_key_values"] = tuple(
                        [
                            (
                                torch.zeros(1, 0, 0, 1, dtype=torch.long, device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                torch.zeros([1, 1, 1, 1], device=input_ids.device).contiguous(),
                                beam_idx_tmp,
                            )
                            for i in range(self.config.num_hidden_layers)
                        ]
                    )
                    has_position_id = False

            if hasattr(self, "trace_graph"):
                if first_token:
                    new_attention_mask = model_inputs["attention_mask"][:batch_size].clone()
                    new_input_ids = model_inputs["input_ids"][:batch_size].clone()
                    if has_position_id:
                        new_position_ids = model_inputs["position_ids"][:batch_size].clone()
                    for i in range(batch_size):
                        new_attention_mask[i] = model_inputs["attention_mask"][i * num_beams]
                        new_input_ids[i] = model_inputs["input_ids"][i * num_beams]
                        if has_position_id:
                            new_position_ids[i] = model_inputs["position_ids"][i * num_beams]
                    model_inputs["attention_mask"] = new_attention_mask
                    model_inputs["input_ids"] = new_input_ids
                    if has_position_id:
                        model_inputs["position_ids"] = new_position_ids
                model_inputs.pop("use_cache", None)
                model_inputs.pop("token_type_ids", None)
                if first_token and hasattr(self, "trace_graph_first"):
                    outputs = self.trace_graph_first(**model_inputs)
                else:
                    outputs = self.trace_graph(**model_inputs)

                if first_token and len(model_inputs["past_key_values"][1]) == 4:
                    outputs = list(outputs)
                    outputs[0] = outputs[0].repeat_interleave(num_beams, dim=0)
                    outputs = tuple(outputs)
                if synced_gpus and this_peer_finished:
                    cur_len = cur_len + 1
                    continue  # don't waste resources running the code we don't need
                next_token_logits = outputs[0][:, -1, :]
            else:
                outputs = self(
                    **model_inputs,
                    return_dict=True,
                    output_attentions=output_attentions,
                    output_hidden_states=output_hidden_states,
                )
                if synced_gpus and this_peer_finished:
                    cur_len = cur_len + 1
                    continue  # don't waste resources running the code we don't need
                next_token_logits = outputs.logits[:, -1, :]
        else:
            outputs = self(
                **model_inputs,
                return_dict=True,
                output_attentions=output_attentions,
                output_hidden_states=output_hidden_states,
            )
            if synced_gpus and this_peer_finished:
                cur_len = cur_len + 1
                continue  # don't waste resources running the code we don't need
            next_token_logits = outputs.logits[:, -1, :]
        next_token_scores = nn.functional.log_softmax(next_token_logits, dim=-1)  # (batch_size * num_beams, vocab_size)
        next_token_scores_processed = logits_processor(input_ids, next_token_scores)
        next_token_scores = next_token_scores_processed + beam_scores[:, None].expand_as(next_token_scores_processed)
        # Store scores, attentions and hidden_states when required
        if return_dict_in_generate:
            if output_scores:
                scores += (next_token_scores_processed,)
            if output_attentions:
                decoder_attentions += (
                    (outputs.decoder_attentions,) if self.config.is_encoder_decoder else (outputs.attentions,)
                )
                if self.config.is_encoder_decoder:
                    cross_attentions += (outputs.cross_attentions,)

            if output_hidden_states:
                decoder_hidden_states += (
                    (outputs.decoder_hidden_states,) if self.config.is_encoder_decoder else (outputs.hidden_states,)
                )

        # reshape for beam search
        vocab_size = next_token_scores.shape[-1]
        next_token_scores = next_token_scores.view(batch_size, num_beams * vocab_size)
        # Sample 1 + len(eos_token_id) next tokens for each beam so we have at least 1 non eos token per beam.
        n_eos_tokens = len(eos_token_id) if eos_token_id else 0
        next_token_scores, next_tokens = torch.topk(
            next_token_scores, max(2, 1 + n_eos_tokens) * num_beams, dim=1, largest=True, sorted=True
        )

        next_indices = torch.div(next_tokens, vocab_size, rounding_mode="floor")
        next_tokens = next_tokens % vocab_size

        # stateless
        beam_outputs = beam_scorer.process(
            input_ids,
            next_token_scores,
            next_tokens,
            next_indices,
            pad_token_id=pad_token_id,
            eos_token_id=eos_token_id,
            beam_indices=beam_indices,
            decoder_prompt_len=decoder_prompt_len,
        )
        beam_scores = beam_outputs["next_beam_scores"]
        beam_next_tokens = beam_outputs["next_beam_tokens"]
        beam_idx = beam_outputs["next_beam_indices"]
        input_ids = torch.cat([input_ids[beam_idx, :], beam_next_tokens.unsqueeze(-1)], dim=-1)
        model_kwargs = self._update_model_kwargs_for_generation(
            outputs, model_kwargs, is_encoder_decoder=self.config.is_encoder_decoder
        )
        if model_kwargs["past_key_values"] is not None:
            model_kwargs["past_key_values"] = self._temporary_reorder_cache(model_kwargs["past_key_values"], beam_idx)

        if return_dict_in_generate and output_scores:
            # pylint: disable=unsubscriptable-object
            beam_indices = tuple((beam_indices[beam_idx[i]] + (beam_idx[i],) for i in range(len(beam_indices))))
        # increase cur_len
        cur_len = cur_len + 1
        if token_latency:
            if input_ids.is_xpu:
                torch.xpu.synchronize()
            latency_list.append(time.time() - tic)

        if beam_scorer.is_done or stopping_criteria(input_ids, scores):
            if not synced_gpus:
                break
            else:
                this_peer_finished = True

    sequence_outputs = beam_scorer.finalize(
        input_ids,
        beam_scores,
        next_tokens,
        next_indices,
        pad_token_id=pad_token_id,
        eos_token_id=eos_token_id,
        max_length=stopping_criteria.max_length,
        beam_indices=beam_indices,
        decoder_prompt_len=decoder_prompt_len,
    )
    if return_dict_in_generate:
        if not output_scores:
            sequence_outputs["sequence_scores"] = None

        if self.config.is_encoder_decoder:
            output_result = BeamSearchEncoderDecoderOutput(
                sequences=sequence_outputs["sequences"],
                sequences_scores=sequence_outputs["sequence_scores"],
                scores=scores,
                beam_indices=sequence_outputs["beam_indices"],
                encoder_attentions=encoder_attentions,
                encoder_hidden_states=encoder_hidden_states,
                decoder_attentions=decoder_attentions,
                cross_attentions=cross_attentions,
                decoder_hidden_states=decoder_hidden_states,
                past_key_values=model_kwargs.get("past_key_values"),
            )
        else:
            output_result = BeamSearchDecoderOnlyOutput(
                sequences=sequence_outputs["sequences"],
                sequences_scores=sequence_outputs["sequence_scores"],
                scores=scores,
                beam_indices=sequence_outputs["beam_indices"],
                attentions=decoder_attentions,
                hidden_states=decoder_hidden_states,
                past_key_values=model_kwargs.get("past_key_values"),
            )
    else:
        output_result = sequence_outputs["sequences"]
    # result
    if token_latency:
        return (output_result, latency_list)
    else:
        return output_result



# Copyright (c) 2025 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import torch

from neural_compressor.torch.algorithms.fp8_quant._core.fp_utils import FP8_143_SCALES, FP8_143_SCALES_TRAITS, calculate_scale_rounding_with_cguid, ScaleCalculationRoundingMode
#TODO [SW-224612]: Use cguid to calc scales and remoce check
from ..common import is_calc_scale_rounding_with_cguid

scale_round_method_registry = {}

def register_scale_round_method(name):
    def decorator(cls):
        scale_round_method_registry[name] = cls
        return cls
    return decorator
 
@register_scale_round_method("POW2")
class ScaleToPow2:
    def __init__(self):
        #TODO [SW-224612]: Use cguid to calc scales and remove check
        if is_calc_scale_rounding_with_cguid():
            self.calc = self.calc_with_cguid

    #TODO [SW-224612]: Use cguid to calc scales and remove special function
    def calc_with_cguid(self, scale):
        scale_pow2 = calculate_scale_rounding_with_cguid(scale, ScaleCalculationRoundingMode.SCALE_TO_POW2_ROUNDING)
        return scale_pow2

    def calc(self, scale):
        scale_pow2 = torch.exp2(torch.ceil(torch.log2(scale)))
        return scale_pow2

@register_scale_round_method("IDENTITY")
class ScaleIdentity:
    def calc(self, scale):
        return scale

@register_scale_round_method("HW_ALIGNED_FIXED")
class ScaleHwAlignedFixed:
    def __init__(self, device_type):
        self.device_type = device_type
    def calc(self, scale):
        hw_aligned_single_scale = FP8_143_SCALES[self.device_type][0]
        return hw_aligned_single_scale

@register_scale_round_method("SCALE_UNIT")
class ScaleUnit:
    def calc(self, scale):
        return 1.0

# Considering range of hw aligned scales: 2^a, 2^a+1,..., 2^b (a<b)
# we want to choose scale s for maxabs m such that 2^a <= s=2^x <= 2^b (for integer a<=x<=b)
# and also 2^(x-1) < m <= 2^x
# if m>=2^b then s=2^b, therefore min(_, 2^b)
# if m<=2^a then s=2^a, therefore max(_, 2^a) --> 2^a <= min(max(_,2^a),2^b) <=2^b
# if s^a<m<2^b then m as a positive number can be written as m=2^y (y=log2(m))
# if y is integer then y=ciel(y) we choose x=y so s=2^x=2^y=2^ciel(y)=2^ciel(log2(m))
# else we choose x=ciel(y) and a<=x-1<y<x<=b and s=2^x=2^ciel(y)=2^ciel(log2(m))
# for Gaudi2 the range is 16^-2..16^1 so we change 2 with 16 and remember that:
# 16 = 2^4, log16(m)=log2(m)/log2(16)=log2(m)/4, and we get:
# we choose s=16^ciel(log16(m))=2^4^ciel(log2(m)/4)=2^(4*ciel(log2(m)/4))=2^(ciel(log2(m)/4)*4)
@register_scale_round_method("HW_ALIGNED")
class ScaleToHwAligned:
    def __init__(self, device_type):
        self.device_type = device_type
    def calc(self, scale):
        scale_pow2 = ScaleToPow2().calc(scale)
        min_scale, max_scale, scale_factor = FP8_143_SCALES_TRAITS[self.device_type]
        scale_pow2_hw = torch.minimum(
            torch.maximum(
                2.0 ** (torch.ceil(torch.log2(scale_pow2) / scale_factor) * scale_factor),
                torch.tensor(min_scale, dtype=scale.dtype, device=scale.device),
                ),
            torch.tensor(max_scale, dtype=scale.dtype, device=scale.device),
        )
        return scale_pow2_hw

